'''Graphical overlays for :class:`~QVideo.lib.QVideoScreen.QVideoScreen`.'''

from .trackpy import QTrackpyOverlay, QTrackpyWidget
from .yolo import QYoloOverlay, QYoloWidget

__all__ = ['QTrackpyOverlay', 'QTrackpyWidget', 'QYoloOverlay', 'QYoloWidget']
